/*
 * ============LICENSE_START=======================================================
 * policy-endpoints
 * ================================================================================
 * Copyright (C) 2017-2019 AT&T Intellectual Property. All rights reserved.
 * Modifications Copyright (C) 2018-2019 Samsung Electronics Co., Ltd.
* ================================================================================
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * ============LICENSE_END=========================================================
 */

package org.onap.policy.common.endpoints.event.comm.bus.internal;

import java.util.UUID;

import org.onap.policy.common.endpoints.event.comm.bus.BusTopicSink;
import org.onap.policy.common.endpoints.utils.NetLoggerUtil;
import org.onap.policy.common.endpoints.utils.NetLoggerUtil.EventType;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Transport Agnostic Bus Topic Sink to carry out the core functionality to interact with a sink
 * regardless if it is UEB or DMaaP.
 *
 */
public abstract class InlineBusTopicSink extends BusTopicBase implements BusTopicSink {

    /**
     * Loggers.
     */
    private static Logger logger = LoggerFactory.getLogger(InlineBusTopicSink.class);

    /**
     * The partition key to publish to.
     */
    protected String partitionId;

    /**
     * Message bus publisher.
     */
    protected BusPublisher publisher;

    /**
     * Constructor for abstract sink.
     * @param busTopicParams contains below listed attributes
     *     servers servers
     *     topic topic
     *     apiKey api secret
     *     apiSecret api secret
     *     partitionId partition id
     *     useHttps does connection use HTTPS?
     *     allowSelfSignedCerts are self-signed certificates allow     *
     * @throws IllegalArgumentException in invalid parameters are passed in
     */
    public InlineBusTopicSink(BusTopicParams busTopicParams) {

        super(busTopicParams);

        if (busTopicParams.isPartitionIdInvalid()) {
            this.partitionId = UUID.randomUUID().toString();
        } else {
            this.partitionId = busTopicParams.getPartitionId();
        }
    }

    /**
     * Initialize the Bus publisher.
     */
    public abstract void init();

    @Override
    public boolean start() {
        logger.info("{}: starting", this);

        synchronized (this) {

            if (this.alive) {
                return true;
            }

            if (locked) {
                throw new IllegalStateException(this + " is locked.");
            }

            this.init();
            this.alive = true;
        }

        return true;
    }

    @Override
    public boolean stop() {

        BusPublisher publisherCopy;
        synchronized (this) {
            this.alive = false;
            publisherCopy = this.publisher;
            this.publisher = null;
        }

        if (publisherCopy != null) {
            try {
                publisherCopy.close();
            } catch (Exception e) {
                logger.warn("{}: cannot stop publisher because of {}", this, e.getMessage(), e);
            }
        } else {
            logger.warn("{}: there is no publisher", this);
            return false;
        }

        return true;
    }

    @Override
    public boolean send(String message) {

        if (message == null || message.isEmpty()) {
            throw new IllegalArgumentException("Message to send is empty");
        }

        if (!this.alive) {
            throw new IllegalStateException(this + " is stopped");
        }

        try {
            synchronized (this) {
                this.recentEvents.add(message);
            }

            NetLoggerUtil.log(EventType.OUT, this.getTopicCommInfrastructure(), this.topic, message);

            publisher.send(this.partitionId, message);
            broadcast(message);
        } catch (Exception e) {
            logger.warn("{}: cannot send because of {}", this, e.getMessage(), e);
            return false;
        }

        return true;
    }

    @Override
    public void setPartitionKey(String partitionKey) {
        this.partitionId = partitionKey;
    }

    @Override
    public String getPartitionKey() {
        return this.partitionId;
    }

    @Override
    public void shutdown() {
        this.stop();
    }

    @Override
    protected boolean anyNullOrEmpty(String... args) {
        for (String arg : args) {
            if (arg == null || arg.isEmpty()) {
                return true;
            }
        }

        return false;
    }

    @Override
    protected boolean allNullOrEmpty(String... args) {
        for (String arg : args) {
            if (!(arg == null || arg.isEmpty())) {
                return false;
            }
        }

        return true;
    }

    @Override
    public String toString() {
        return "InlineBusTopicSink [partitionId=" + partitionId + ", alive=" + alive + ", publisher=" + publisher + "]";
    }
}
